/*
  This is the mpc_pk component of the Amanzi code. 

  Copyright 2010-201x held jointly by LANS/LANL, LBNL, and PNNL. 
  Amanzi is released under the three-clause BSD License. 
  The terms of use and "as is" disclaimer for this license are 
  provided in the top-level COPYRIGHT file.

*/

#include "mpc_coupled_transport.hh"
#include "Transport_PK_ATS.hh"

namespace Amanzi {

  CoupledTransport_PK::CoupledTransport_PK(Teuchos::ParameterList& pk_tree_or_fe_list,
                                           const Teuchos::RCP<Teuchos::ParameterList>& global_list,
                                           const Teuchos::RCP<State>& S,
                                           const Teuchos::RCP<TreeVector>& soln) :
    PK(pk_tree_or_fe_list, global_list, S, soln),
    WeakMPC(pk_tree_or_fe_list, global_list, S, soln)
  {

    //std::cout << *global_list<<"\n";
    //exit(0);
      // Create verbosity object.
    vo_ = Teuchos::null;
    Teuchos::ParameterList vlist;
    vlist.sublist("verbose object") = global_list -> sublist("verbose object");
    vo_ =  Teuchos::rcp(new VerboseObject("Coupled TransportPK", vlist)); 
    
    name_ = "surface subsurface transport";

    Teuchos::Array<std::string> pk_order = plist_->get< Teuchos::Array<std::string> >("PKs order");
  
    for (int i=0; i<2; i++){
      Teuchos::RCP<Teuchos::ParameterList> list = Teuchos::sublist(Teuchos::sublist(global_list, "PKs"), pk_order[i]);      
      Key dom_name = list->get<std::string>("domain name", "domain");
      if (dom_name == "domain"){
        subsurface_transport_list_ = list;
        subsurface_name_ = dom_name;
        mesh_ = S->GetMesh(subsurface_name_);
        subsurf_id_ = i;
      }else{
        surface_transport_list_ = list;
        surface_name_ = dom_name;
        surf_mesh_ = S->GetMesh(surface_name_);
        surf_id_ = i;
      }
    }

    subsurface_flux_key_ =  plist_->get<std::string>("flux_key", 
                                                     Keys::getKey(subsurface_name_, "mass_flux"));

    surface_flux_key_ =  plist_->get<std::string>("flux_key", 
                                                  Keys::getKey(surface_name_, "mass_flux"));


  }

// -----------------------------------------------------------------------------
// Calculate the min of sub PKs timestep sizes.
// -----------------------------------------------------------------------------
double CoupledTransport_PK::get_dt() {
  //double dt = Amanzi::PK_MPCSubcycled_ATS::get_dt();
  //ComputeVolumeDarcyFlux(S_next_.ptr());

  double surf_dt = sub_pks_[surf_id_]->get_dt();
  double subsurf_dt = sub_pks_[subsurf_id_]->get_dt();

  Teuchos::OSTab tab = vo_->getOSTab();
  if (vo_->os_OK(Teuchos::VERB_HIGH)) {
    *vo_->os()<< "surface transport dt = "<<surf_dt<<"\n";
    *vo_->os()<< "sub surface transport dt = "<<subsurf_dt<<"\n";
  }

  double dt = std::min(surf_dt, subsurf_dt);
  set_dt(dt);
  return dt;
 
}


// -----------------------------------------------------------------------------
// Make necessary operatios by the end of the time steps.
// -----------------------------------------------------------------------------
// void CoupledTransport_PK::CommitStep(double t_old, double t_new, const Teuchos::RCP<State>& S) {
//   sub_pks_[master_]->CommitStep(t_old, t_new, S);
//   sub_pks_[slave_]->CommitStep(t_old, t_new, S);
// }



void CoupledTransport_PK::Setup(const Teuchos::Ptr<State>& S){

  //passwd_ = "coupled_transport";  // owner's password
  //passwd_ = "state";  // owner's password

  WeakMPC::Setup(S);
  
}

void CoupledTransport_PK::Initialize(const Teuchos::Ptr<State>& S){

  WeakMPC::Initialize(S);

}

// void CoupledTransport_PK::ComputeVolumeDarcyFlux(const Teuchos::Ptr<State>& S){

//   int  nfaces_owned;

//   Teuchos::RCP<const Epetra_MultiVector> darcy_flux = 
//     S->GetFieldData(Keys::getKey(subsurface_name_,"mass_flux"))->ViewComponent("face");

//   Teuchos::RCP<const Epetra_MultiVector> surf_darcy_flux =
//     S->GetFieldData(Keys::getKey(surface_name_,"mass_flux"))->ViewComponent("face");

//   Key molar_den_key = Keys::getKey(subsurface_name_, "molar_density_liquid");
//   Teuchos::RCP<const Epetra_MultiVector> molar_density = 
//     S->GetFieldData(molar_den_key)->ViewComponent("cell", true);

//   Key surf_molar_den_key = Keys::getKey(surface_name_, "molar_density_liquid");
//   Teuchos::RCP<const Epetra_MultiVector> surf_molar_density = 
//     S->GetFieldData(surf_molar_den_key)->ViewComponent("cell", true);


//   // subsurface
//   nfaces_owned = mesh_->num_entities(AmanziMesh::FACE, AmanziMesh::Parallel_type::OWNED);
//   AmanziMesh::Entity_ID_List cells;
  
//   for (int f = 0; f < nfaces_owned; f++){
//     mesh_->face_get_cells(f, AmanziMesh::Parallel_type::ALL, &cells);
//     double n_liq=0.;
//     for (int c=0; c<cells.size();c++) n_liq += (*molar_density)[0][c];
//     n_liq /= cells.size();
//     if (n_liq > 0) (*vol_darcy)[0][f] = (*darcy_flux)[0][f]/n_liq;
//     else (*vol_darcy)[0][f] = 0.;
//   }
//   S->GetField(vol_darcy_key_, passwd_)->set_initialized();  

//   // surface
//   nfaces_owned = surf_mesh_->num_entities(AmanziMesh::FACE, AmanziMesh::Parallel_type::OWNED);
//   for (int f = 0; f < nfaces_owned; f++){
//     mesh_->face_get_cells(f, AmanziMesh::Parallel_type::ALL, &cells);
//     double n_liq=0.;
//     for (int c=0; c<cells.size();c++) n_liq += (*surf_molar_density)[0][c];
//     n_liq /= cells.size();
//     if (n_liq > 0) (*surf_vol_darcy)[0][f] = (*surf_darcy_flux)[0][f]/n_liq;
//     else (*surf_vol_darcy)[0][f] = 0.;
//   }
//   S->GetField(surf_vol_darcy_key_, passwd_)->set_initialized();
    

// }


// -----------------------------------------------------------------------------
// Advance each sub-PK individually, returning a failure as soon as possible.
// -----------------------------------------------------------------------------
bool CoupledTransport_PK::AdvanceStep(double t_old, double t_new, bool reinit) {
  bool fail = false;

  //double dt_MPC = S_->final_time() - S_->initial_time();


  sub_pks_[subsurf_id_]->AdvanceStep(t_old, t_new, reinit);
  *vo_->os() <<"Subsurface step successful\n";
  sub_pks_[surf_id_]->AdvanceStep(t_old, t_new, reinit);
  *vo_->os() <<"Overland step successful\n";


  const Epetra_MultiVector& surf_tcc = *S_inter_->GetFieldCopyData("surface-total_component_concentration", "subcycling")->ViewComponent("cell",false);  
  const Epetra_MultiVector& tcc = *S_inter_->GetFieldCopyData("total_component_concentration", "subcycling")->ViewComponent("cell",false);

  const std::vector<std::string>&  component_names_sub =
    Teuchos::rcp_dynamic_cast<Transport::Transport_PK_ATS>(sub_pks_[subsurf_id_])->component_names();

  int num_components =  Teuchos::rcp_dynamic_cast<Transport::Transport_PK_ATS>(sub_pks_[subsurf_id_]) -> num_aqueous_component();
   std::vector<double> mass_subsurface(num_components, 0.), mass_surface(num_components, 0.);

  
  if (vo_->getVerbLevel() >= Teuchos::VERB_MEDIUM){
    for (int i=0; i<num_components; i++){
      mass_subsurface[i] = Teuchos::rcp_dynamic_cast<Transport::Transport_PK_ATS>(sub_pks_[subsurf_id_])
        ->ComputeSolute(tcc, i);
      mass_surface[i] = Teuchos::rcp_dynamic_cast<Transport::Transport_PK_ATS>(sub_pks_[surf_id_])
        ->ComputeSolute(surf_tcc, i);
      Teuchos::OSTab tab = vo_->getOSTab();

      *vo_->os() <<"subsurface =" << mass_subsurface[i] << " mol";
      *vo_->os() <<", surface =" << mass_surface[i]<< " mol";
      *vo_->os() <<", ToTaL =" << mass_surface[i]+mass_subsurface[i]<< " mol" <<std::endl;
    }
  }
  
  return fail;
  

}


/* *******************************************************************
* Interpolate linearly in time between two values v0 and v1. The time 
* is measuared relative to value v0; so that v1 is at time dt. The
* interpolated data are at time dt_int.            
******************************************************************* */
void CoupledTransport_PK::InterpolateCellVector(
    const Epetra_MultiVector& v0, const Epetra_MultiVector& v1, 
    double dt_int, double dt, Epetra_MultiVector& v_int) 
{
  double a = dt_int / dt;
  double b = 1.0 - a;
  v_int.Update(b, v0, a, v1, 0.);
}



}  // namespace Amanzi
